# standard imports
import logging
import tempfile
import os
import sys

# external imports
from mimeparse import parse_mime_type

# local imports
from .base import Renderer as BaseRenderer
from .base import stream_accumulator

logg = logging.getLogger(__name__)

def to_suffixed_file(d, s, data):
    (v, ext) = os.path.splitext(s)
    r = tempfile.mkstemp(suffix=ext, dir=d)

    f = os.fdopen(r[0], 'wb')
    try:
        f.write(data)
    except TypeError:
        f.write(data.encode('utf-8'))
    f.close()

    return r[1]


class Accumulator:

    def __init__(self, w=sys.stdout):
        self.w = w


    def add(self, v):
        stream_accumulator(v, w=self.w)


class Renderer(BaseRenderer):

    def __init__(self, basket, dump_dir=None, accumulator=None, **kwargs):
        if accumulator == None:
            accumulator = Accumulator().add
        super(Renderer, self).__init__(basket, accumulator=accumulator, **kwargs)
        self.dump_dir = dump_dir


    def apply_issue(self, state, issue, tags, accumulator=None):
        if self.render_mode == 0:
            return self.apply_issue_standalone(state, issue, tags, accumulator=accumulator)

        s = '{}\t{}\t{}'.format(
                issue.title,
                ','.join(tags),
                issue.id,
                )
        if issue.alias != None:
            s += " (" + issue.alias + ")"
        s += "\n"
        self.add(s, accumulator=accumulator)
        super(Renderer, self).apply_issue(state, issue, tags, accumulator=accumulator)


    def apply_issue_standalone(self, state, issue, tags, accumulator=None):
        s = """title: {}
id: {}
state: {}
tags: {}
""".format(
        issue.title,
        issue.id,
        state,
        ', '.join(tags),
            )
        self.add(s, accumulator=accumulator)

        assigned = issue.get_assigned()

        assigns = [] 
        s = 'assigned to: '
        if len(assigned) == 0:
            assigns.append('(not assigned)')
        else:
            owner = issue.owner()
            for v in assigned:
                o = v[0]
                ss = o.id()
                if o == owner:
                    ss += ' (owner)'
                #s = '\t' + str(s) + '\n'
                assigns.append(ss)
        s += ', '.join(assigns) + '\n'

        if len(issue.dependencies) > 0:
            s += 'depends on: '
            s += ', '.join(issue.dependencies) + '\n'
        self.add(s, accumulator=accumulator)

        super(Renderer, self).apply_issue(state, issue, tags, accumulator=accumulator)


    def apply_message(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        s = '\nmessage {} from {} {} - {}\n'.format(
            message_date,
            envelope.sender,
            envelope.valid,
            message_id,
            )
        return s


    def apply_message_part(self, state, issue, tags, envelope, message, message_date, message_content, accumulator=None):
        if message_content['filename'] != None:
            if self.dump_dir != None:
                filename = to_suffixed_file(self.dump_dir, message_content['filename'], message_content['contents'])
            sz = message_content['size']
            if sz == -1:
                sz = 'unknown'
            v = '[file: {}, type {}/{}, size: {}]'.format(
                    message_content['filename'],
                    message_content['type'][0],
                    message_content['type'][1],
                    sz,
                    )
        else:
            v = message_content['contents']

        s = '\n\t' + v + '\n'
        return s


    def apply_state(self, state, accumulator=None):
        s = '[' + state + ']\n'
        self.add(s, accumulator=accumulator) 
        super(Renderer, self).apply_state(state, accumulator=accumulator)


    def apply_state_post(self, state, accumulator=None):
        self.add('\n', accumulator=accumulator)



# standard imports
import logging
import uuid
import mimetypes
from base64 import b64encode
import time
from email.message import Message
from email import message_from_string
from email.policy import Compat32
from email.utils import formatdate
from email.utils import parsedate_to_datetime


logg = logging.getLogger(__name__)


class MessageEnvelope:

    def __init__(self, msg):
        self.msg = msg
        self.sender = None
        self.valid = False
        self.resolved = False


def rubber_stamp_envelope(envelope, envelope_type):
    envelope.valid = True
    return


class IssueMessage:

    def __init__(self, issue):
        self.__m = Message()

        self.__m.add_header('Subject', issue.title)
        self.__m.add_header('X-Piknik-Id', issue.id)
        self.__m.add_header('Date', formatdate(time.time()))
        self.__m.set_payload(None)
        self.__m.set_type('multipart/relative')
        self.__m.set_boundary(str(uuid.uuid4()))


    def __unwrap(self, msg, envelope_callback=rubber_stamp_envelope, message_callback=None, post_callback=None):
        message_ids = []
        message_id = None
        message_date = None
        envelope = None
        initial = False

        for m in msg.walk():
            if not initial:
                initial = True
                continue
            env_header = m.get('X-Piknik-Envelope')
            if env_header != None:
                if envelope_callback != None:
                    m = envelope_callback(m, env_header)
                envelope = m
                continue

            if message_callback == None:
                continue

            new_message_id = m.get('X-Piknik-Msg-Id')
            if new_message_id != None:
                message_id = new_message_id
                message_ids.append(message_id)
                d = m.get('Date')
                message_date = parsedate_to_datetime(d)
            message_callback(m, message_id, message_date)

        if post_callback != None:
            post_callback(message_ids)

        return message_ids


    @classmethod
    def parse(cls, issue, v, envelope_callback=None, message_callback=None, post_callback=None):
        o = cls(issue)
        m = message_from_string(v)
        o.__unwrap(m, envelope_callback=envelope_callback, message_callback=message_callback, post_callback=post_callback)
        o.__m = m
        return o


    def from_text(self, v):
        m = Message()
        m.add_header('Content-Transfer-Encoding', 'QUOTED-PRINTABLE')
        m.set_charset('UTF-8')
        m.set_payload(str(v))
        return m


    def detect_file(self, v):
        r = mimetypes.guess_type(v)
        if r[0] == None:
            return ('application/octet-stream', None,)
        return r


    def from_file(self, v):
        m = Message()
        
        mime_type = self.detect_file(v)
        m.set_type(mime_type[0])

        if mime_type[1] != None:
            m.set_charset(mime-type[1])
        m.add_header('Content-Transfer-Encoding', 'BASE64')
        m.add_header('Content-Disposition', 'attachment; filename="{}"'.format(v))

        f = open(v, 'rb')
        r = f.read()
        f.close()
        r = b64encode(r)
        m.add_header('Content-Length', str(len(r)))
        m.set_payload(r.decode())

        return m


    def add(self, *args, related_id=None, wrapper=None, message_id=None):
        m_id = None
        try:
            m_id = uuid.UUID(message_id)
        except (ValueError, TypeError):
            m_id = uuid.uuid4()
        m = Message()
        m.add_header('X-Piknik-Msg-Id', str(m_id))
        m.add_header('Date', formatdate(time.time()))
        if related_id != None:
            m.add_header('In-Reply-To', related_id)
        m.set_payload(None)
        m.set_type('multipart/mixed')
        m.set_boundary(str(uuid.uuid4()))
        for a in args:
            p = a[:2]
            v = a[2:]
            r = None
            if p == 'f:':
                r = self.from_file(v)
            elif p == 's:':
                r = self.from_text(v)
            m.attach(r)
        if wrapper:
            m = wrapper(m)
        self.__m.attach(m)


    def as_string(self, **kwargs):
        return self.__m.as_string(**kwargs)


    def as_bytes(self, **kwargs):
        return self.__m.as_bytes(**kwargs)


    def __str__(self):
        return self.as_string()



# standard imports
import os

# local imports
from piknik import Basket
from piknik.store import FileStoreFactory
from piknik.crypto import PGPSigner

class Context:

    def __init__(self, arg, assembler, mode=0, gpg_home=os.environ.get('GPGHOME'), create=True):
        self.issue_id = arg.issue_id
        self.files_dir = arg.files_dir
        self.alias = getattr(arg, 'alias', None)
        self.show_finished = getattr(arg, 'show_finished', False)
        self.show_states = getattr(arg, 'state', [])
        #self.store_factory = FileStoreFactory(arg.d)
        store_factory = FileStoreFactory(arg.d, create=create)
        self.signer = None
        sign_fn = None
        if hasattr(arg, 's'):
            self.signer = PGPSigner(default_key=arg.s, use_agent=True)
            sign_fn = self.signer.sign
        self.basket = Basket(store_factory, message_wrapper=sign_fn)
        self.gpg_home = gpg_home
        assembler(self, arg)



# standard imports
import uuid
import json
import datetime

# local imports
from piknik.identity import Identity
from piknik.error import UnknownIdentityError
from piknik.error import ExistsError


def id_generator():
        return str(uuid.uuid4())

class Issue:

    def __init__(self, title, issue_id=None, alias=None, id_generator=id_generator):
        if issue_id == None:
            issue_id = id_generator()
        self.id = issue_id
        self.title = title
        self.assigned = []
        self.assigned_time = []
        self.dependencies = []
        self.alias = alias
        self.owner_idx = 0


    @staticmethod
    def from_str(s):
        r = json.loads(s)
        o = Issue(title=r['title'], issue_id=r['id'])
        for i, k in enumerate(r['assigned'].keys()):
            p = Identity(k)
            o.assigned.append(p)
            t = datetime.datetime.utcfromtimestamp(r['assigned'][k])
            o.assigned_time.append(t)
            if r['owner'] == None or k == r['owner']:
                r['owner'] = k
                o.owner_idx = i
        for v in r['dependencies']:
            o.dep(v)
        o.alias = r.get('alias')
        return o


    def assign(self, identity, t=None):
        if identity in self.assigned:
            raise ExistsError(identity)
        if t == None:
            t = datetime.datetime.utcnow()
        self.assigned.append(identity)
        self.assigned_time.append(t)

    
    def dep(self, dependency_issue_id):
        if dependency_issue_id in self.dependencies:
            raise ExistsError(self.id, dependency_issue_id)
        self.dependencies.append(dependency_issue_id)


    def undep(self, dependency_issue_id=None):
        if dependency_issue_id == None:
            self.dependencies = []
            return
        self.dependencies.remove(dependency_issue_id)


    def get_assigned(self):
        return list(zip(self.assigned, self.assigned_time))


    def get_dependencies(self):
        return self.dependencies


    def unassign(self, identity):
        for i, v in enumerate(self.assigned):
            if v == identity:
                self.assigned.remove(v)
                if i == self.owner_idx:
                    self.owner_idx = 0
                return True
        raise UnknownIdentityError(identity)


    def owner(self):
        try:
            return self.assigned[self.owner_idx]
        except IndexError:
            pass

        raise UnknownIdentityError


    def set_owner(self, identity):
        r = self.owner()
        if identity == r:
            return False

        for i, v in enumerate(self.assigned):
            if v == identity:
                self.owner_idx = i
                return True

        raise UnknownIdentityError(identity)

    
    def __str__(self):
        o = {
            'id': str(self.id),
            'title': self.title,
            'assigned': {},
            'dependencies': self.dependencies,
            'alias': self.alias,
            'owner': None,
            }

        for i, v in enumerate(self.get_assigned()):
            aid = v[0].id()
            o['assigned'][aid] = v[1].timestamp()
            if self.owner_idx == i:
                o['owner'] = aid

        return json.dumps(o)


    def __eq__(self, o):
        if o.id != self.id:
            return False
        if o.title != self.title:
            return False
        if len(self.assigned) != len(o.assigned):
            return False
        for i, v in enumerate(self.assigned):
            if o.assigned[i] != v:
                return False
            
        return True



# standard imports
import io
import os
import sys
import argparse
import logging
import tempfile
from base64 import b64decode
from email.utils import parsedate_to_datetime
import importlib
import copy

# local imports
from piknik.cli import Context
from piknik.cli.show import subparser as subparser_show
from piknik.cli.mod import subparser as subparser_mod
from piknik.cli.add import subparser as subparser_add
from piknik.cli.comment import subparser as subparser_comment

logging.basicConfig(level=logging.WARNING)
logg = logging.getLogger()

argp = argparse.ArgumentParser()
argp.add_argument('-d', type=str, help='Data directory')
argp.add_argument('-f', '--files', dest='f', action='store_true', help='Save attachments to filesystem')
argp.add_argument('-o', '--files-dir', dest='files_dir', type=str, help='Directory to output saved files to')
argp.add_argument('-v', action='store_true', help='Turn on debug logs')
argp.add_argument('-i','--issue-id',  type=str, help='Issue id to show')
argp.add_argument('--alias', type=str, help='alias string to refer to issue to with cli commands')
argp.add_argument('cmd', type=str, nargs='?', choices=['show', 'add', 'mod', 'comment'], help='subcommand to execute')
strargs = copy.copy(sys.argv[1:])

have_help = False
try:
    strargs.remove('-h')
    have_help = True
except ValueError:
    pass
try:
    strargs.remove('--help')
    have_help = True
except ValueError:
    pass
arg, unknown = argp.parse_known_args(strargs)

m = None
create = False
if arg.cmd == 'show':
    m = importlib.import_module('piknik.cli.show')
elif arg.cmd == 'add':
    m = importlib.import_module('piknik.cli.add')
    create = True
elif arg.cmd == 'mod':
    m = importlib.import_module('piknik.cli.mod')
elif arg.cmd == 'comment':
    m = importlib.import_module('piknik.cli.comment')
else:
    if not have_help:
        #raise ValueError('unknown subcommand')
        m = importlib.import_module('piknik.cli.show')

if m != None:
    argp = m.subparser(argp)

arg = argp.parse_args(sys.argv[1:])
if arg.v:
    logg.setLevel(logging.DEBUG)

m.ctx = Context(arg, m.assembler, create=create)


def main():
    global m
    m.main()


if __name__ == '__main__':
    main()



# standard imports
import sys
import logging

logg = logging.getLogger(__name__)

ctx = None


def subparser(argp):
    argp.add_argument('-s', '--sign-as', dest='s', type=str, help='PGP fingerprint of key to sign issue update with')
    argp.add_argument('-x', '--text', dest='x', type=str, help='Add literal message text')
    argp.add_argument('-y', '--file', dest='y', type=str, action='append', default=[], help='Add arbitrary file as content')
    return argp


def assembler(o, arg):
    o.x = arg.x
    o.y = arg.y
    o.s = arg.s


next_i = 1
def next_message_arg():
    global next_i

    r = sys.argv[next_i]

    if r[0] != '-':
        next_i += 1
        return None

    if r[1] not in ['r', 'i', 'x', 'y']:
        next_i += 1
        return None

    v = sys.argv[next_i+1]
    next_i += 2
    return (r[1], v,)


def main():
    messages = []
    while True:
        try:
            r = next_message_arg()
        except IndexError:
            break
        if r == None:
            continue

        if r[0] == 'x':
            messages.append('s:' + r[1])
        elif r[0] == 'y':
            messages.append('f:' + r[1])

    ctx.basket.msg(ctx.issue_id, *messages)




# standard imports
import logging
import sys

# local imports
from piknik.msg import MessageEnvelope

logg = logging.getLogger(__name__)


def stream_accumulator(v, w=sys.stdout):
    w.write(v)


class Renderer:

    def __init__(self, basket, accumulator=None, wrapper=None, states_include=[], states_skip=[]):
        self.b = basket
        self.a = accumulator
        self.w = wrapper
        if self.w == None:
            logg.info('no wrapper defined. no message parts will be output')
        self.render_mode = 0
        self.states = []
        self.__make_state_filter(states_include, states_skip)
       

    def __make_state_filter(self, include, skip):
        have_include = False
        for i in range(len(include)):
            include[i] = include[i].upper()
            have_include = True
        for i in range(len(skip)):
            skip[i] = skip[i].upper()
        for state in self.b.states():
            if have_include:
                if state not in include:
                    continue
            elif state in skip:
                continue
            self.states.append(state)


    def add(self, v, accumulator=None):
        if accumulator == None:
            accumulator = self.a
        if accumulator != None:
            if v != None:
                r = accumulator(v)


    def apply_envelope_pre(self, state, issue, tags, envelope, accumulator=None):
        pass


    def apply_envelope_post(self, state, issue, tags, envelope, accumulator=None):
        pass

    
    def apply_envelope(self, state, issue, tags, envelope, accumulator=None):
        pass


    def apply_message_pre(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        pass


    def apply_message_post(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        pass

 
    def apply_message_post(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        pass

   
    def apply_message(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        pass


    def apply_message_part(self, state, issue, tags, envelope, message, message_date, message_content, accumulator=None):
        pass


    def apply_issue_pre(self, state, issue, tags, accumulator=None):
        pass


    def apply_issue_post(self, state, issue, tags, accumulator=None):
        pass


    def apply_issue_mid(self, state, issue, tags, accumulator=None):
        pass


    def apply_issue(self, state, issue, tags, accumulator=None):

        def envelope_callback(envelope, envelope_type):
            if self.w != None:
                envelope = self.w.process_envelope(envelope, envelope_type)
            else:
                envelope = MessageEnvelope(envelope)
            r = self.apply_envelope_pre(state, issue, tags, envelope, accumulator=accumulator)
            self.add(r)
            r = self.apply_envelope(state, issue, tags, envelope, accumulator=accumulator)
            self.add(r)
            r = self.apply_envelope_post(state, issue, tags, envelope, accumulator=accumulator)
            self.add(r)
            return envelope

        def message_callback(message, message_id, message_date):
            envelope = None
            if self.w == None:
                return (envelope, message,)

            (envelope, message) = self.w.process_message(message, message_id, message_date)

            initial = True
            while True:
                v = self.w.pop()
                if v == None:
                    break
                if initial:
                    r = self.apply_message_pre(state, issue, tags, envelope, message, message_id, message_date, accumulator=accumulator)
                    self.add(r)
                    r = self.apply_message(state, issue, tags, envelope, message, message_id, message_date, accumulator=accumulator)
                    self.add(r)
                    r = self.apply_message_post(state, issue, tags, envelope, message, message_id, message_date, accumulator=accumulator)
                    self.add(r)
                    initial = False

                r = self.apply_message_part(state, issue, tags, envelope, message, message_date, v)
                self.add(r)

            return (envelope, message,)

        self.b.get_msg(issue.id, envelope_callback=envelope_callback, message_callback=message_callback)


    def apply_state_pre(self, state, accumulator=None):
        self.render_mode = 1


    def apply_state_post(self, state, accumulator=None):
        pass


    def apply_state(self, state, accumulator=None):
        for issue_id in self.b.list(category=state):
            issue = self.b.get(issue_id)
            tags = self.b.tags(issue_id=issue_id)
            r = self.apply_issue_pre(state, issue, tags)
            self.add(r)
            r = self.apply_issue(state, issue, tags)
            self.add(r)
            r = self.apply_issue_post(state, issue, tags)
            self.add(r)


    def apply_begin(self, accumulator=None):
        pass


    def apply_end(self, accumulator=None):
        pass


    def apply(self, accumulator=None):
        r = self.apply_begin()
        self.add(r)

        #for state in self.b.states():
        for state in self.states:
            r = self.apply_state_pre(state)
            self.add(r)
            r = self.apply_state(state)
            self.add(r)
            r = self.apply_state_post(state)
            self.add(r)

        r = self.apply_end()
        self.add(r)



# standard imports
import os
import logging
import importlib
import argparse

# local imports
from piknik import Issue
from piknik.crypto import PGPSigner

ctx = None
accumulator_f = None
accumulator = None

logg = logging.getLogger(__name__)



def set_accumulator(ctx, m, issue_id=None):
    global accumulator_f
    global accumulator
    if ctx.files_dir != None:
        fb = None
        if issue_id == None:
            fb = 'index.html'
        else:
            fb = issue_id + '.html'
        fp = os.path.join(ctx.files_dir, fb)
        accumulator_f = open(fp, 'w')
        accumulator = m.Accumulator(w=accumulator_f)
        return accumulator.add
    return None


def reset_accumulator():
    global accumulator_f
    global accumulator
    if accumulator_f != None:
        accumulator_f.close()
        accumulator_f = None
        issues = accumulator.issues
        accumulator = None
        return issues
    return []


def subparser(argp):
    argp.add_argument('-r', '--renderer', type=str, default='default', help='Renderer module for output')
    argp.add_argument('-s', '--state', type=str, action='append', default=[], help='Limit results to state(s)')
    argp.add_argument('-k', '--trusted-key', dest='trusted_key',  type=str, action='append', default=[], help='Key to trust, even if revoked')
    argp.add_argument('--show-finished', dest='show_finished', action='store_true', help='Include finished issues')
    argp.add_argument('--reverse', action='store_true', help='Sort comments by oldest first')
    return argp


def assembler(o, arg):
    o.renderer = arg.renderer
    o.trust = []
    for v in arg.trusted_key:
        l = len(v)
        if l < 8:
            raise ValueError('Minimum 4 bytes match on trust key')
        elif int(l) not in [8, 16, 40]:
            raise ValueError('Invalid fingerprint length')
        o.trust.append(v.upper())


def main():
    global ctx

    renderer_s = ctx.renderer
    if renderer_s == 'default':
        renderer_s = 'piknik.render.plain'
    elif renderer_s == 'html':
        renderer_s = 'piknik.render.html'

    m = None
    try:
        m = importlib.import_module(renderer_s)
    except ModuleNotFoundError:
        renderer_s = 'piknik.render.' + renderer_s
        m = importlib.import_module(renderer_s)

    accumulator = None
    accumulator_f = None

    issues = []
    if ctx.issue_id:
        issues.append(ctx.issue_id)

    states_skip = []
    if not ctx.show_finished:
        states_skip.append('FINISHED')

    if ctx.issue_id == None:
        accumulator = set_accumulator(ctx, m)
        renderer = m.Renderer(ctx.basket, accumulator=accumulator, states_include=ctx.show_states, states_skip=states_skip)
        renderer.apply()
        issues = reset_accumulator()

    for issue_id in issues:
        accumulator = set_accumulator(ctx, m, issue_id=issue_id)
        issue = ctx.basket.get(issue_id)
        tags = ctx.basket.tags(issue_id)
        state = ctx.basket.get_state(issue_id)
        verifier = PGPSigner(home_dir=ctx.gpg_home, skip_verify=False, keytrust=ctx.trust)
        renderer = m.Renderer(ctx.basket, wrapper=verifier, accumulator=accumulator, states_include=ctx.show_states, states_skip=states_skip)

        renderer.apply_begin()
        renderer.apply_issue_pre(state, issue, tags)
        renderer.apply_issue(state, issue, tags)
        renderer.apply_issue_post(state, issue, tags)
        renderer.apply_end()
        
        reset_accumulator()



# standard imports
import os
import logging
import tempfile
from email.message import Message

# external imports
import gnupg

# local imports
from piknik.error import VerifyError
from piknik.error import UnknownIdentityError
from piknik.error import SignError
from piknik.wrap import Wrapper

logg = logging.getLogger(__name__)
logging.getLogger('gnupg').setLevel(logging.ERROR)


class PGPSigner(Wrapper):

    def __init__(self, home_dir=None, dump_dir=None, default_key=None, passphrase=None, use_agent=False, skip_verify=False, keytrust=[]):
        super(PGPSigner, self).__init__(dump_dir=dump_dir)
        self.gpg = gnupg.GPG(gnupghome=home_dir)
        self.default_key = default_key
        self.passphrase = passphrase
        self.use_agent = use_agent
        self.sign_material = None
        self.pre_buffer = []
        self.keytrust = keytrust
        self.__skip_verify = skip_verify


    def set_from(self, msg, passphrase=None):
        r = None
        for v in self.gpg.list_keys(True):
            if self.default_key == None or v['fingerprint'].upper() == self.default_key.upper():
                r = v['uids'][0]
                break
        if r == None:
            raise UnknownIdentityError('no signing keys found')
        msg.add_header('From', r)


    def sign(self, msg, passphrase=None): # msg = IssueMessage object
        m = Message()
        m.set_type('multipart/relative')
        m.add_header('X-Piknik-Envelope', 'pgp')
        ms = Message()
        ms.set_type('application/pgp-signature')
        fn = '{}.asc'.format(msg.get('X-Piknik-Msg-Id'))
        ms.add_header('Content-Disposition', 'attachment', filename=fn)

        if passphrase == None:
            passphrase = self.passphrase
        self.set_from(msg, passphrase=passphrase)
        v = msg.as_string()
        sig = self.gpg.sign(v, keyid=self.default_key, detach=True, passphrase=passphrase)
        if sig.status != 'signature created':
            raise SignError()
        ms.set_payload(str(sig))
    
        m.attach(msg)
        m.attach(ms)

        return m

    
    def process_envelope(self, msg, env_header):
        self.envelope = None
        if env_header != 'pgp':
            raise VerifyError('expected envelope type "pgp", but got {}'.format(env_header))
        if self.envelope_state > -1 and self.envelope_state < 2:
            raise VerifyError('new envelope before previous was verified ({})'.format(self.envelope_state))
        self.sign_material = None
        super(PGPSigner, self).process_envelope(msg, env_header)
        return self.envelope


    def process_message(self, msg, message_id, message_date):
        if msg.get('From') != None:
            self.envelope.sender = msg.get('From')

        if self.envelope_state == 0:
            self.envelope_state = 1
            self.sign_material = msg
            return (self.envelope, msg,)

        if msg.get('Content-Type') != 'application/pgp-signature':
            self.pre_buffer.append((self.envelope, message_id, msg,))
            return (self.envelope, msg,)

        v = self.sign_material.as_string()

        sig = msg.get_payload()
        (fd, fp) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w')
        f.write(sig)
        f.close()
        r = self.gpg.verify_data(fp, v.encode('utf-8'))
        os.unlink(fp)
        
        if r.key_status != None:
            if 'revoked' in r.key_status:
                have = False
                for k in self.keytrust:
                    key_used_part = r.key_id[(len(k)*-1):]
                    logg.debug('check used revoked key {} matchpart {} against {}'.format(r.key_id, key_used_part, k))
                    if key_used_part == k:
                        have = True
                        break
                if not have:
                    raise VerifyError('revoked key {} not in accept list'.format(r.key_id))
            else:
                raise VerifyError('unexpected key status {}'.format(r.key_status))
        if r.status == 'no public key':
            logg.warning('public key for {} not found, cannot verify'.format(r.fingerprint))
        elif r.status != 'signature valid':
            if self.__skip_verify:
                logg.warning('invalid signature for message {}'.format(message_id))
            else:
                raise VerifyError('invalid signature for message {}'.format(message_id))
        else:
            logg.debug('signature ok from {}'.format(r.fingerprint))
            self.envelope.valid = True
        if self.envelope.sender == None:
            self.envelope.sender = r.fingerprint
        self.envelope_state = 2
        
        while True:
            try:
                v = self.pre_buffer.pop(0)
                self.add(v[0], v[1], v[2])
            except IndexError:
                break

        return (self.envelope, msg,)



# standard imports
import os
import unittest
import logging
import json
import shutil
from email.message import Message

# local imports
from piknik import Basket
from piknik import Issue
from piknik.msg import IssueMessage
from piknik.msg import MessageEnvelope

# test imports
from tests.common import TestStates
from tests.common import TestMsgStore
from tests.common import pgp_setup

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()

test_dir = os.path.realpath(os.path.dirname(__file__))


class TestMsg(unittest.TestCase):

    def setUp(self):
        self.store = TestStates()
        (self.crypto, self.gpg, self.gpg_dir) = pgp_setup()
        self.b = Basket(self.store, message_wrapper=self.crypto.sign)


    def tearDown(self):
        shutil.rmtree(self.gpg_dir)


    def test_wrap_sig(self):
        m = Message()
        m.set_type('multipart/mixed')
        m.set_payload(None)

        one = Message()
        one.set_charset('utf-8')
        one.set_payload('foo')
        m.attach(one)

        two = Message()
        two.set_charset('utf-8')
        two.set_payload('bar')
        m.attach(two)

        o = Issue('foo')
        m = self.crypto.sign(m, passphrase='foo')
        self.crypto.envelope = MessageEnvelope(m)
        self.crypto.envelope_state = 0
        r = IssueMessage.parse(o, str(m), envelope_callback=self.crypto.process_envelope, message_callback=self.crypto.process_message)


    def test_wrap_double_sig(self):
        mp = Message()
        mp.set_type('multipart/related')
        mp.set_payload(None)

        m = Message()
        m.add_header('X-Piknik-Msg-Id', 'foo')
        m.set_type('multipart/mixed')
        m.set_payload(None)

        one = Message()
        one.set_charset('utf-8')
        one.set_payload('inky')
        m.attach(one)

        two = Message()
        two.set_charset('utf-8')
        two.set_payload('pinky')
        m.attach(two)

        m = self.crypto.sign(m, passphrase='foo')
        mp.attach(m)

        m = Message()
        m.add_header('X-Piknik-Msg-Id', 'bar')
        m.set_type('multipart/mixed')
        m.set_payload(None)

        one = Message()
        one.set_charset('utf-8')
        one.set_payload('blinky')
        m.attach(one)

        two = Message()
        two.set_charset('utf-8')
        two.set_payload('clyde')
        m.attach(two)

        m = self.crypto.sign(m, passphrase='foo')
        mp.attach(m)

        self.crypto.process_envelope(mp, 'pgp')
        r = self.crypto.process_message(mp, m, 'foo')
        r = self.crypto.process_message(mp, m, 'bar')


    # TODO: assert
    def test_wrap_basket_sig(self):
        o = Issue('foo')
        v = self.b.add(o)
        r = self.b.msg(v, 's:foo', 's:bar')
        print(r)


    def test_wrap_basket_sig(self):
        o = Issue('foo')
        v = self.b.add(o)
        r = self.b.msg(v, 's:foo', 's:bar')


if __name__ == '__main__':
    unittest.main()



import hashlib


class Identity:
    
    def __init__(self, fingerprint):
        self.__id = bytes.fromhex(fingerprint)
       

    def id(self):
        return self.__id.hex()


    def __eq__(self, other):
        return other.id() == self.id()

       
    def uri(self):
        h = hashlib.sha256()
        h.update(self.__id)
        z = h.digest()
        return 'sha256:' + z.hex()


    def __str__(self):
        return self.id()



class DeadIssue(Exception):
    pass


class ExistsError(Exception):
    pass


class UnknownIdentityError(Exception):
    pass


class VerifyError(Exception):
    pass


class SignError(Exception):
    pass



# standard imports
import unittest
import logging
import json

# local imports
from piknik import Issue
from piknik.identity import Identity

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


class TestIssue(unittest.TestCase):

    def test_basic(self):
        o = Issue('foo')
        r = json.loads(str(o))
        self.assertEqual(str(o.id), r['id'])
        self.assertEqual(o.title, r['title'])


    def test_basic_from_str(self):
        o = Issue('foo')
        v = str(o)
        r = Issue.from_str(v)
        self.assertTrue(o == r)


if __name__ == '__main__':
    unittest.main()



# standard imports
import unittest
import logging
import json
import shutil
import io
import tempfile
from email.message import Message
from email.utils import localtime as email_localtime

# external imports
from mimeparse import parse_mime_type

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.msg import IssueMessage
from piknik.render.base import Renderer
from piknik.wrap import Wrapper

# test imports
from tests.common import TestStates
from tests.common import TestMsgStore
from tests.common import pgp_setup

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


def test_wrapper(p):
    m = Message()
    m.add_header('Foo', 'bar')
    m.set_type('multipart/relative')
    m.add_header('X-Piknik-Envelope', 'foo')
    m.set_payload(p)
    return m


def test_unwrapper(msg, message_callback=None, part_callback=None):
    for v in msg.walk():
        if message_callback != None:
            message_callback(v)


class TestRenderer(Renderer):

    def __init__(self, basket, accumulator=None):
        super(TestRenderer, self).__init__(basket, accumulator=accumulator, wrapper=Wrapper())
        self.p = 0
        self.e = 0


    def apply_envelope(self, state, issue, tags, envelope, accumulator=None):
        r = self.e
        self.e += 1
        return r


    def apply_message_part(self, state, issue, tags, envelope, message, message_date, message_content, accumulator=None):
        r = self.p
        self.p += 1
        return r


class TestRendererComposite(TestRenderer):

    def __init__(self, basket, accumulator=None):
        super(TestRendererComposite, self).__init__(basket, accumulator=accumulator)
        self.last_message_id = None
        self.m = []


    def apply_message_post(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        if self.last_message_id != message_id:
            self.m.append(message_id)
            self.last_message_id = message_id


class TestMsg(unittest.TestCase):

    def setUp(self):
        self.acc = []
        self.store = TestStates()
        self.b = Basket(self.store, message_wrapper=test_wrapper)
        self.render_dir = tempfile.mkdtemp()


    def accumulate(self, v):
        self.acc.append(v)


    def tearDown(self):
        shutil.rmtree(self.render_dir)


    def test_idlepass(self):
        wrapper = Wrapper()
        renderer = TestRenderer(self.b, accumulator=self.accumulate)
        issue_one = Issue('foo')
        self.b.add(issue_one)

        issue_two = Issue('bar')
        v = self.b.add(issue_two)

        m = self.b.msg(v, 's:foo')

        renderer.apply()
        self.assertEqual(len(self.acc), 2)
        self.assertEqual(renderer.e, 1)
        self.assertEqual(renderer.p, 1)


    def test_composite(self):
        renderer = TestRendererComposite(self.b, accumulator=self.accumulate)
        issue_one = Issue('foo')
        self.b.add(issue_one)

        issue_two = Issue('bar')
        v = self.b.add(issue_two)

        m = self.b.msg(v, 's:foo')

        renderer.apply()
        self.assertEqual(len(renderer.m), 1)


    def test_attachment(self):
        renderer = TestRendererComposite(self.b, accumulator=self.accumulate)
        issue_one = Issue('foo')
        self.b.add(issue_one)

        issue_two = Issue('bar')
        v = self.b.add(issue_two)

        m = self.b.msg(v, 'f:tests/one.png')

        renderer.apply()
        self.assertEqual(len(renderer.m), 1)


if __name__ == '__main__':
    unittest.main()



# standard imports
import unittest
import logging
import tempfile
import shutil
# external imports
import shep

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.error import DeadIssue
from piknik.store import FileStoreFactory
from piknik.identity import Identity

# tests imports
from tests.common import debug_out
from tests.common import TestStates
from tests.common import pgp_setup


logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


def debug_out(self, k, v):
    logg.debug('TRACE: {} {}'.format(k, v))


class TestStore(unittest.TestCase):

    def setUp(self):
        self.d = tempfile.mkdtemp()
        logg.debug('tempdir is {}'.format(self.d))
        self.store_factory = FileStoreFactory(self.d)
        self.b = Basket(self.store_factory)


    def tearDown(self):
        shutil.rmtree(self.d)
        pass


    # TODO: assert
    def test_basic(self):
        o = Issue('foo')
        v = self.b.add(o)


    # TODO: assert
    def test_load(self):
        o = Issue('foo')
        va = self.b.add(o)

        o = Issue('bar')
        vb = self.b.add(o)
        
        self.b.advance(va)

        b = Basket(self.store_factory)
        r = b.get(va)


    def test_load_tag(self):
        o = Issue('foo')
        va = self.b.add(o)
        self.b.tag(va, 'inky')
        self.b.tag(va, 'pinky')

        b = Basket(self.store_factory)
        r = b.tags(va)
        self.assertIn('INKY', r)
        self.assertIn('PINKY', r)

        self.b.untag(va, 'inky')
        b = Basket(self.store_factory)
        r = b.tags(va)
        self.assertNotIn('INKY', r)
        self.assertIn('PINKY', r)


    # TODO: assert
    def test_msg_putget(self):
        o = Issue('foo')
        issue_id = self.b.add(o)
        m = self.b.msg(issue_id, 's:bar')


    # TODO: assert
    def test_msg_resume(self):
        o = Issue('foo')
        v = self.b.add(o)
        m = self.b.msg(v, 's:bar')

        b = Basket(self.store_factory)
        m = b.msg(v, 's:baz')


    def test_msg_sig_verify_resume(self):
        (crypto, gpg, gpg_dir) = pgp_setup()
        b = Basket(self.store_factory, message_wrapper=crypto.sign)
        o = Issue('foo')
        v = b.add(o)
        r = b.msg(v, 's:foo', 's:bar')

        b = Basket(self.store_factory, message_wrapper=crypto.sign)
        m = b.get_msg(v)


    def test_store_assignments(self):
        o = Issue('foo')
        vb = self.b.add(o)
   
        alice = b'inky'.hex()
        bob = b'pinky'.hex()

        self.b.assign(vb, alice)
        self.b.assign(vb, bob)
        self.b.owner(vb, bob)
        r = self.b.get(vb)
        check = Identity(bob)
        self.assertTrue(r.owner() == check)

        self.b.unassign(vb, alice)

        r = self.b.get(vb)
        check = Identity(bob)
        self.assertEqual(r.owner(), check)
        check_assigned = r.get_assigned()
        self.assertTrue(r.owner() == check)


if __name__ == '__main__':
    unittest.main()



# standard imports
import unittest
import logging

# external imports
import shep

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.error import DeadIssue

# tests imports
from tests.common import debug_out
from tests.common import TestStates

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


def debug_out(self, k, v):
    logg.debug('TRACE: {} {}'.format(k, v))


class TestBasic(unittest.TestCase):

    def setUp(self):
        self.b = Basket(TestStates())


    def test_setget(self):
        o = Issue('foo')
        v = self.b.add(o)
        self.b.tag(v, 'inky')
        self.b.tag(v, 'pinky')
        self.b.untag(v, 'pinky')


    def test_setget_alias(self):
        o = Issue('foo', alias='bar')
        v = self.b.add(o)
        self.b.tag('bar', 'inky')
        self.b.tag('bar', 'pinky')
        self.b.untag('bar', 'pinky')
        self.assertEqual(self.b.tags('bar'), ['INKY']) 


 
if __name__ == '__main__':
    unittest.main()



# standard imports
import sys

# local imports
from piknik import Issue

ctx = None


def subparser(argp):
    argp.add_argument('title', type=str, nargs='*', help='issue title')
    argp.add_argument('--proposed', type=bool, help='start issue in proposed state')
    return argp


def assembler(o, arg):
    o.title = arg.title
    o.proposed = arg.proposed


def main():
    title = ''
    for s in ctx.title:
        if s == ' ':
            continue
        if title != '':
            title += ' '
        title += s
    if ctx.alias == None:
        issue_id = ctx.basket.id_generator()
        o = Issue(title, issue_id=issue_id, alias=issue_id[:5])
    else:
        o = Issue(title, alias=ctx.alias)

    v = ctx.basket.add(o)

    if not ctx.proposed:
        ctx.basket.state_backlog(issue_id)

    sys.stdout.write(v + '\n')



# standard imports
import unittest
import logging
import json

# local imports
from piknik import Issue
from piknik.identity import Identity
from piknik.error import UnknownIdentityError

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


class TestAssign(unittest.TestCase):

    def setUp(self):
        self.alice = 'F3FAF668E82EF5124D5187BAEF26F4682343F692'
        self.bob = 'F645E047EE5BC4E2824C94DB42DC91CFA8ABA02B'


    def test_identity_pointer(self):
        check = "sha256:65ea9b0609341322903823660bf45326f473d903ee7d327dff442f46d68eacd9"
        p = Identity(self.alice)
        r = p.uri()
        self.assertEqual(r, check)


    def test_identity_load(self):
        o = Issue('foo')
        alice = Identity(self.alice)
        o.assign(alice)
        bob = Identity(self.bob)
        o.assign(bob)
        r = o.get_assigned()
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0][0], alice)
        self.assertEqual(r[1][0], bob)
        self.assertGreater(r[1][1], r[0][1])


    def test_identity_assigned_from_str(self):
        o = Issue('foo')
        alice = Identity(self.alice)
        bob = Identity(self.bob)
        o.assign(alice)
        o.assign(bob)
        v = str(o)
        r = Issue.from_str(v)
        self.assertTrue(o == r)

        check = r.get_assigned()
        self.assertEqual(len(check), 2)


    def test_identity_set_owner(self):
        o = Issue('foo')
        alice = Identity(self.alice)
        bob = Identity(self.bob)
        o.assign(alice)
        o.assign(bob)

        r = o.owner()
        self.assertEqual(r, alice)

        o.set_owner(bob)
        r = o.owner()
        self.assertEqual(r, bob)


    def test_identity_unassign(self):
        o = Issue('foo')
        alice = Identity(self.alice)
        bob = Identity(self.bob)
        o.assign(alice)
        o.assign(bob)
   
        o.unassign(alice)
        r = o.get_assigned()
        self.assertEqual(len(r), 1)

        r = o.owner()
        self.assertEqual(r, bob)

        with self.assertRaises(UnknownIdentityError):
            o.unassign(alice)

        o.unassign(bob)
        with self.assertRaises(UnknownIdentityError):
            o.owner()


    def test_issue_identity_equality(self):
        alice = Identity(self.alice)
        bob = Identity(self.bob)

        one = Issue('foo')
        two = Issue('foo', issue_id=one.id)

        one.assign(alice)
        two.assign(alice)
        self.assertTrue(one == two)

        two.assign(bob)
        self.assertFalse(one == two)
        

if __name__ == '__main__':
    unittest.main()



# standard imports
import os
import uuid
import logging

# external imports
from shep.store.file import SimpleFileStoreFactory
from shep.persist import PersistedState
from leveldir.hex import HexDir

logg = logging.getLogger(__name__)


def default_formatter(hx):
    return hx.lower()


class MsgDir(HexDir):

    def __init__(self, root_path):
        super(MsgDir, self).__init__(root_path, 16, levels=2, prefix_length=0, formatter=default_formatter)


    def __check(self, key, content, prefix):
        pass


    def get(self, k):
        u = uuid.UUID(k)
        fp = self.to_filepath(u.hex)
        f = open(fp, 'rb')
        r = f.read()
        f.close()
        return r


    def key_to_string(self, k):
        u = uuid.UUID(bytes=k)
        return u.bytes.hex()


    def put(self, k, v):
        u = uuid.UUID(k)
        logg.debug('putting {}'.format(u.bytes.hex()))
        return self.add(u.bytes, v)


class AliasDir:

    def __init__(self, root_path):
        self.dir = root_path
        os.makedirs(self.dir, exist_ok=True)


    def get(self, k):
        fp = os.path.join(self.dir, k)
        f = open(fp, 'rb')
        r = f.read()
        f.close()
        u = uuid.UUID(r.hex())
        return str(u)


    def key_to_string(self, k):
        u = uuid.UUID(bytes=k)
        return u.bytes.hex()


    def put(self, k, v):
        u = uuid.UUID(v)
        fp = os.path.join(self.dir, k)
        logg.debug('putting {} {}'.format(u.bytes.hex(), fp))
        f = open(fp, 'xb')
        f.write(u.bytes)
        f.close()


    def purge(self, k):
        fp = os.path.join(self.dir, k)
        os.remove(fp)


class FileStoreFactory:

    def __init__(self, directory=None, create=True):
        if directory == None:
            directory = os.path.join('.', '.piknik')
        self.directory = directory
        if not create:
            if not os.path.exists(self.directory):
                raise FileNotFoundError(self.directory)


    def create_states(self, logger=None, default_state=None, verifier=None):
        factory = SimpleFileStoreFactory(self.directory).add
        return PersistedState(factory, 6, logger=logger, verifier=verifier, default_state=default_state)

    
    def create_tags(self, logger=None, default_state=None, verifier=None):
        directory = os.path.join(self.directory, '.tags')
        os.makedirs(directory, exist_ok=True)
        factory = SimpleFileStoreFactory(directory)
        state = PersistedState(factory.add, 0, logger=logger, check_alias=False, default_state='untagged')
        aliases = []
        for k in factory.ls():
            if k == 'UNTAGGED':
                continue
            elif k[0] == '_':
                aliases.append(k)
                continue
            state.add(k)

        for v in aliases:
            s = state.from_elements(v, create_missing=True)
            state.alias(v, s)
                
        return state


    def create_messages(self):
        d = os.path.join(self.directory, '.msg')
        return MsgDir(d)


    def create_aliases(self):
        d = os.path.join(self.directory, '.alias')
        return AliasDir(d)



# standard imports
import unittest
import logging
import json

# local imports
from piknik import Basket
from piknik import Issue
from piknik.identity import Identity
from piknik.error import UnknownIdentityError
from piknik.error import ExistsError

# tests imports
from tests.common import TestStates

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


class TestAssign(unittest.TestCase):

    def setUp(self):
        self.alice = 'F3FAF668E82EF5124D5187BAEF26F4682343F692'
        self.bob = 'F645E047EE5BC4E2824C94DB42DC91CFA8ABA02B'


    def test_dep_basic(self):
        one = Issue('foo')
        one.dep('bar')
        one.dep('baz')
        with self.assertRaises(ExistsError):
            one.dep('bar')
        one.undep('bar')
        self.assertEqual(len(one.dependencies), 1)


    def test_dep_alias(self):
        one = Issue('foo', alias='inky')
        two = Issue('bar', alias='pinky')
        three = Issue('baz')
        self.b = Basket(TestStates())
        issue_id_one = self.b.add(one)
        issue_id_two = self.b.add(two)
        issue_id_three = self.b.add(three)
        self.b.dep('inky', 'pinky')
        self.b.dep(issue_id_three, 'inky')

        with self.assertRaises(ExistsError):
            self.b.dep(issue_id_one, 'pinky')

        with self.assertRaises(ExistsError):
            self.b.dep(issue_id_three, issue_id_one)


if __name__ == '__main__':
    unittest.main()



# standard imports
import logging
from base64 import b64decode

# external imports
from mimeparse import parse_mime_type

# local imports
from piknik.msg import MessageEnvelope

logg = logging.getLogger(__name__)


class Wrapper:

    def __init__(self, dump_dir=None):
        self.content_buffer = []
        self.envelope_state = -1 # -1 not in envelope, 0 in outer envelope, 1 inner envelope, not (yet) valid, 2 envelope valid (with signature)
        self.envelope = None
        self.dump_dir = dump_dir
        self.msg_idx = 0


    def process_envelope(self, msg, env_header):
        self.envelope = MessageEnvelope(msg)
        self.envelope_state = 0
        self.msg_idx = 0
        return self.envelope


    def process_message(self, msg, message_id, message_date):
        if msg.get('X-Piknik-Msg-Id') == None:
            self.add(self.envelope, message_id, msg)
        return (self.envelope, msg,)


    def add(self, envelope, message_id, message):
        m = parse_mime_type(message.get_content_type())
        filename = message.get_filename()

        v = None
        if filename == None:
            if m[0] == 'text':
                if m[1] == 'plain':
                    v = message.get_payload()
                    if message.get('Content-Transfer-Encoding') == 'BASE64':
                        v = b64decode(v).decode()
        else:
            v = message.get_payload()
            if message.get('Content-Transfer-Encoding') == 'BASE64':
                try:
                    v = b64decode(v).decode()
                except UnicodeDecodeError:
                    pass

            if self.dump_dir != None:
                (void, ext) = os.path.splitext(filename)
                (fp, fn) = tempfile.mkstemp(suffix=ext, dir=self.dump_dir)

                f = os.fdopen(fp, 'wb')
                try:
                    f.write(v)
                except TypeError:
                    f.write(v.encode('utf-8'))
                f.close()

                filename = fn

        sz = message.get('Content-Length')
        if sz == None:
            sz = -1
        else:
            sz = int(sz)

        o = {
                'envelope': envelope,
                'id': message_id,
                'type': m,
                'filename': filename,
                'size': sz,
                'contents': v,
                'idx': self.msg_idx,
                }
        self.msg_idx += 1

        self.content_buffer.append(o)


    def pop(self):
        try:
            return self.content_buffer.pop(0)
        except IndexError:
            return None



# standard imports
import logging

# external imports
import shep
from shep.error import StateItemNotFound

# local imports
from .error import DeadIssue
from .issue import Issue
from .issue import id_generator
from .msg import IssueMessage
from .identity import Identity

logg = logging.getLogger(__name__)

class Basket:

    def __init__(self, state_factory, message_wrapper=None):
        self.no_resurrect = True
        self.state = state_factory.create_states(default_state='proposed', verifier=self.__check_resurrect)
        self.state.add('backlog')
        self.state.add('pending')
        self.state.add('doing')
        self.state.add('review')
        self.state.add('finished')
        self.state.add('blocked')
        self.state.alias('doingblocked', self.state.DOING | self.state.BLOCKED)
        self.state.alias('pendingblocked', self.state.PENDING | self.state.BLOCKED)
        self.limit = self.state.FINISHED
        self.state.sync()

        self.__tags = state_factory.create_tags()
        self.__tags.sync(ignore_auto=False)

        self.__msg = state_factory.create_messages()
        self.__msg_wrap = message_wrapper

        self.__alias = state_factory.create_aliases()

        self.issues_rev = {}


    def id_generator(self):
        while True:
            issue_id = id_generator()
            issue_alias = issue_id[:5]
            try:
                self.__alias.get(issue_alias)
            except FileNotFoundError:
                return issue_id


    def __check_resurrect(self, st, k, f, t):
        if self.no_resurrect:
            if f & self.state.FINISHED > 0:
                raise DeadIssue(k)


    def add(self, issue):
        issue_id = str(issue.id)
        j = str(issue)
        if issue.alias != None:
            self.__alias.put(issue.alias, issue.id)
        self.state.put(issue_id, contents=j)
        self.__tags.put(issue_id)
        return issue_id


    def get(self, issue_id):
        logg.debug("basket issue get {}".format(issue_id))
        r = self.state.get(issue_id)
        if r == None:
            aliased_issue_id = self.__alias.get(issue_id)
            r = self.state.get(aliased_issue_id)
            logg.debug("resolved issue {} from alias '{}': {}".format(aliased_issue_id, issue_id, r))
        o = Issue.from_str(r)
        return o


    def get_state(self, issue_id):
        o = self.get(issue_id)
        v = self.state.state(o.id)
        return self.state.name(v)


    def list(self, category=None):
        if category == None:
            category = self.state.BACKLOG
        else:
            category = self.state.from_name(category)
        return self.state.list(category)


    def state_pending(self, issue_id):
        o = self.get(issue_id)
        self.state.move(o.id, self.state.PENDING)


    def state_doing(self, issue_id):
        o = self.get(issue_id)
        self.state.move(o.id, self.state.DOING)


    def state_review(self, issue_id):
        o = self.get(issue_id)
        self.state.move(o.id, self.state.REVIEW)


    def state_backlog(self, issue_id):
        o = self.get(issue_id)
        self.state.move(o.id, self.state.BACKLOG)


    def state_finish(self, issue_id):
        o = self.get(issue_id)
        self.state.move(o.id, self.state.FINISHED)
        if o.alias != None:
            self.__alias.purge(o.alias)


    def advance(self, issue_id):
        o = self.get(issue_id)
        if self.state.state(o.id) & self.limit > 0:
            raise DeadIssue(o.id)
        self.unblock(o.id)
        self.state.next(o.id)


    def unblock(self, issue_id):
        o = self.get(issue_id)
        if self.state.state(o.id) & self.state.BLOCKED > 0:
            self.state.unset(o.id, self.state.BLOCKED)


    def block(self, issue_id):
        o = self.get(issue_id)
        self.state.set(o.id, self.state.BLOCKED)


    def blocked(self):
        return self.list('blocked')


    def states(self):
        return self.state.all(pure=True, bit_order=True)


    def tag(self, issue_id, tag):
        v = 0
        try:
            v = self.__tags.from_name(tag)
        except AttributeError:
            self.__tags.add(tag)
            v = self.__tags.from_name(tag)
     
        o = self.get(issue_id)
        move = False
        try:
            r = self.__tags.state(o.id)
            if r == 0:
                move = True
        except shep.error.StateItemNotFound:
            self.__tags.put(o.id)
            move = True

        if move:
            self.__tags.move(o.id, v)
        else:
            self.__tags.set(o.id, v)



    def untag(self, issue_id, tag):
        v = self.__tags.from_name(tag)
        o = self.get(issue_id)
        self.__tags.unset(o.id, v, allow_base=True)


    def tags(self, issue_id):
        o = self.get(issue_id)
        v = None
        try:
            v = self.__tags.state(o.id)
        except StateItemNotFound:
            self.__tags.put(o.id)
            #self.__tags.set(o.id, self.__tags.UNTAGGED)
            v = self.__tags.state(o.id)
        r = self.__tags.elements(v)
        if r == 'UNTAGGED':
            r = '(' + r + ')'
        return shep.state.split_elements(r)


    def __get_msg(self, issue_id, envelope_callback=None, message_callback=None, post_callback=None):
        o = self.get(issue_id)
        try:
            v = self.__msg.get(o.id)
            m = IssueMessage.parse(o, v.decode('utf-8'), envelope_callback=envelope_callback, message_callback=message_callback, post_callback=post_callback)
            return m
        except FileNotFoundError as e:
            logg.debug('instantiating new message log for {} {}'.format(o.id, e))

        return IssueMessage(o)


    def get_msg(self, issue_id, envelope_callback=None, message_callback=None, post_callback=None):
        return self.__get_msg(issue_id, envelope_callback=envelope_callback, message_callback=message_callback, post_callback=post_callback)
 
    
    def dep(self, issue_id, dependency_issue_id):
        o = self.get(issue_id)
        od = self.get(dependency_issue_id)
        r = o.dep(od.id)
        self.state.replace(o.id, contents=str(o))
        return r


    def undep(self, issue_id, dependency_issue_id):
        o = self.get(issue_id)
        self.get(dependency_issue_id)
        r = o.undep(dependency_issue_id)
        self.state.replace(o.id, contents=str(o))
        return r


    def assign(self, issue_id, identity):
        o = self.get(issue_id)
        v = Identity(identity)
        r = o.assign(v)
        self.state.replace(o.id, contents=str(o))
        return r


    def owner(self, issue_id, identity):
        o = self.get(issue_id)
        v = Identity(identity)
        r = o.set_owner(v)
        self.state.replace(o.id, contents=str(o))
        return r


    def unassign(self, issue_id, identity):
        o = self.get(issue_id)
        v = Identity(identity)
        r = o.unassign(v)
        self.state.replace(o.id, contents=str(o))
        return r


    def msg(self, issue_id, *args):
        o = self.get(issue_id)
        m = self.__get_msg(o.id)
        m.add(*args, wrapper=self.__msg_wrap)
        ms = m.as_bytes()
        self.__msg.put(o.id, ms)
        return m



# standard imports
import sys
import os
import logging

# external imports
import dominate
from dominate.tags import div, p, a, meta, ul, ol, li, h1, h2, link, dl, dd, dt, img
from mimeparse import parse_mime_type

# local imports
from .base import Renderer as BaseRenderer

logg = logging.getLogger(__name__)


class Accumulator:

    def __init__(self, w=sys.stdout):
        self.doc = None
        self.category = ul(_id='state_list')
        self.category_content = None
        self.issue = None
        self.msg = None
        #self.envelope_content = None
        self.envelope = None
        self.w = w
        self.last_v = None
        self.issues = []


    def add(self, v, w=None):
        if w == None:
            w = self.w
        if len(v) == 0:
            if self.envelope != None:
                self.msg.add(self.envelope)
            if self.msg != None:
                self.category.add(self.msg)
                #self.doc.add(self.msg)
            if self.last_v != None:
                self.category.add(li(self.last_v))
            self.doc.add(self.category)
            w.write(self.doc.render())
            self.last_v = None

        v_id = getattr(v, 'id', '')
        logg.debug('add id {}'.format(v_id))
        if len(v_id) > 1:
            if v_id[0] == 's':
                self.last_v = v
            if v_id[:2] == 's_':
                if self.issue != None:
                    self.category_content.add(self.issue)
                    self.category.add(li(self.category_content))
                self.category_content = v
                self.issue = ul(_id='issue_list_' + v_id[2:])
            elif v_id[:2] == 'i_':
                logg.debug('issue now')
                self.issue.add(v)
                self.issues.append(v_id[2:])
            elif v_id[:4] == 'd_i_':
                self.category = v
                self.issue = v
                self.msg = ol(_id='message_list')
                self.issues.append(v_id[2:])
            elif v_id[:4] == 'd_c_':
                self.issue.add(v)
            elif v_id[:2] == 'm_':
                if self.envelope != None:
                    self.msg.add(self.envelope)
                #self.envelope_content = v
                self.envelope = li(v)
            elif v_id[:4] == 'd_m_':
                self.envelope.add(v)
        else:
            self.doc = v
            self.last_v = None


class Renderer(BaseRenderer):

    def __init__(self, basket, accumulator=None, wrapper=None, outdir=None, states_include=[], states_skip=[]):
        if accumulator == None:
            accumulator = Accumulator().add
        super(Renderer, self).__init__(basket, accumulator=accumulator, wrapper=wrapper, states_include=states_include, states_skip=states_skip)
        self.outdir = outdir


    def apply_state(self, state, accumulator=None):
        self.render_mode = 1
        v = div(_id='s_' + state.lower())
        v.add(h2(state))
        self.add(v)
        super(Renderer, self).apply_state(state, accumulator=accumulator)


    def apply_issue(self, state, issue, tags, accumulator=None):
        if self.render_mode == 1:
            s = issue.title
            u = a(s, href=issue.id + '.html')
            return li(u, _id='i_' + issue.id)

        r = div(_id='d_i_' + issue.id)

        s = h1(issue.title)
        r.add(s)
        
        s = dd()
        s.add(dt('issue id'))
        s.add(dd(issue.id))

        s.add(dt('state'))
        s.add(dd(state))

        s.add(dt('tags'))
        if len(tags) == 1 and tags[0] == '(UNTAGGED)':
                s.add(dd(tags[0]))
        else:
            r_d = ul()
            for v in tags:
                r_d.add(li(v))
            s.add(dd(r_d))
    
        assigned = issue.get_assigned()
        s.add(dt('assigned to'))
        if len(assigned) == 0:
            s.add(dd('(not assigned)'))
        else:
            owner = issue.owner()
            r_r = ul()
            for v in assigned:
                o = v[0]
                ss = o.id()
                if o == owner:
                    ss += ' (owner)'
                r_r.add(li(ss))
            s.add(dd(r_r))

        #r.add(s)

        deps = issue.get_dependencies()
        s.add(dt('depends on'))
        if len(deps) == 0:
            s.add(dd('no dependencies'))
        else:
            r_r = ul()
            for v in deps:
                r_r.add(dd(v))
            s.add(dd(r_r))
    
        r.add(s)

        self.add(r)

        r = self.apply_issue_mid(state, issue, tags, accumulator=accumulator)
        self.add(r)

        super(Renderer, self).apply_issue(state, issue, tags, accumulator=accumulator)

    
    def apply_message(self, state, issue, tags, envelope, message, message_id, message_date, accumulator=None):
        r = div(_id='m_' + message_id)

        s = dd()
        s.add(dt('Date'))
        s.add(dd(str(message_date)))

        v = envelope.sender
        if v == None:
            v = '(unknown)'
        else:
            if envelope.valid:
                v += ' (!!)'
            else:
                v += ' (??)'
        s.add(dt('By'))
        s.add(dd(v))
        r.add(s)

        self.add(r)
       

    def apply_message_part(self, state, issue, tags, envelope, message, message_date, message_content, accumulator=None):
        logg.debug('env {}'.format(envelope.sender))
        r = None
        if message_content['filename'] != None:
            s = 'data:{}/{};base64,{}'.format(
                    message_content['type'][0],
                    message_content['type'][1],
                    message_content['contents'],
                    )
            r = None
            if message_content['type'][0] == 'image':
                r = img(src=s)
            else:
                v = os.path.basename(message_content['filename'])
                r = a(v, href=s)
            
        else:
            r = message_content['contents']

        m_id = 'd_m_{}_{}'.format(
                    message_content['id'],
                    message_content['idx'],
                    )

        return div(r, _id=m_id)


    def apply_begin(self, accumulator=None):
        r = dominate.document(title='issues for ...')
        r.head.add(meta(name='generator', content='piknik'))
        r.head.add(link(rel='stylesheet', href='style.css'))
        self.add(r)


    def apply_end(self, accumulator=None):
        self.add(())
        return None


    def apply_issue_mid(self, state, issue, tags, accumulator=None):
        logg.debug('>>>>>>>> foo')
        return h2('comments', _id='d_c_{}'.format(issue.id))



# standard imports
import unittest
import logging

# external imports
import shep

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.error import DeadIssue

# tests imports
from tests.common import debug_out
from tests.common import TestStates

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


class TestBasic(unittest.TestCase):

    def setUp(self):
        self.b = Basket(TestStates())

    
    def test_issue_basic(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        self.assertEqual(v, str(o.id))
        r = self.b.get(v)
        self.assertEqual(r, o)


    def test_list(self):
        o = Issue('The first issue')
        self.b.add(o)
        o = Issue('The second issue')
        self.b.add(o)
        r = self.b.list('proposed')
        self.assertEqual(len(r), 2)


    def test_progres(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        self.b.advance(v)
        self.b.advance(v)
        self.b.advance(v)
        self.b.advance(v)
        self.b.advance(v)
        with self.assertRaises(DeadIssue):
            self.b.advance(v)
            

    def test_list_jump(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        o_two = Issue('The second issue')
        v_two = self.b.add(o_two)
        self.b.advance(v_two)
        o_three = Issue('The second issue')
        self.b.add(o_three)
        self.b.state_doing(v)

        r = self.b.list('proposed')
        self.assertEqual(len(r), 1)

        r = self.b.list('backlog')
        self.assertEqual(len(r), 1)

        r = self.b.list('doing')
        self.assertEqual(len(r), 1)


    def test_jump(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        self.b.state_doing(v)
        r = self.b.list('doing')
        self.assertEqual(len(r), 1)
        self.b.state_review(v)
        r = self.b.list('review')
        self.assertEqual(len(r), 1)
        self.b.state_backlog(v)
        r = self.b.list('backlog')
        self.assertEqual(len(r), 1)
        self.b.state_finish(v)
        r = self.b.list('finished')
        self.assertEqual(len(r), 1)


    def test_magic_unblock(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        self.b.advance(v)
        self.b.advance(v)
        self.b.block(v)
        self.assertIn(v, self.b.blocked())
        self.b.advance(v)
        self.assertNotIn(v, self.b.blocked())


    def test_no_resurrect(self):
        o = Issue('The first issue')
        v = self.b.add(o)
        self.b.state_finish(v)
        with self.assertRaises(DeadIssue):
            self.b.state_doing(v)


    def test_states_list(self):
        r = self.b.states()
        self.assertEqual(len(r), 7)


    def test_alias(self):
        o = Issue('The first issue', alias="first")
        issue_id = o.id
        v = self.b.add(o)
        o = self.b.get("first")
        self.assertEqual(o.id, issue_id)
        self.b.state_finish(issue_id)
        with self.assertRaises(FileNotFoundError):
            o = self.b.get("first")

if __name__ == '__main__':
    unittest.main()



# standard imports
import logging

logg = logging.getLogger(__name__)

ctx = None


def subparser(argp):
    argp.add_argument('--accept', action='store_true', help='Accept proposed issue')
    argp.add_argument('--block', action='store_true', help='Set issue as blocked')
    argp.add_argument('--unblock', action='store_true', help='Set issue as unblocked')
    argp.add_argument('--finish', action='store_true', help='Set issue as finished (alias of -s finish)')
    argp.add_argument('-s', '--state', type=str, help='Move to state')
    argp.add_argument('-t', '--tag', type=str, action='append', default=[], help='Add tag to issue')
    argp.add_argument('-u', '--untag', type=str, action='append', default=[], help='Remove tag from issue')
    #argp.add_argument('-f', '--file', type=str, action='append', help='Add message file part')
    #argp.add_argument('-m', '--message', type=str, action='append', default=[], help='Add message text part')
    argp.add_argument('-a', '--assign', type=str, action='append', default=[], help='Assign given identity to issue')
    argp.add_argument('--unassign', type=str, action='append', default=[], help='Unassign given identity from issue')
    argp.add_argument('--owner', type=str, help='Set given identity as owner of issue')
    argp.add_argument('--dep', action='append', default=[], type=str, help='Set issue dependency')
    argp.add_argument('--undep', action='append', default=[], type=str, help='Remove issue dependency')
    return argp


def assembler(o, arg):
    o.owner = arg.owner
    o.block = arg.block
    o.unblock = arg.unblock
    o.state = arg.state
    o.finish = arg.finish
    o.accept = arg.accept
    o.tag = arg.tag
    o.untag = arg.untag
    o.assign = arg.assign
    o.unassign = arg.unassign
    o.dep = arg.dep
    o.undep = arg.undep


def main():
    global ctx

    o = ctx.basket.get(ctx.issue_id)

    if ctx.block:
        ctx.basket.block(ctx.issue_id)
    elif ctx.unblock:
        ctx.basket.unblock(ctx.issue_id)

    if ctx.state != None:
        m = getattr(ctx.basket, 'state_' + ctx.state)
        m(ctx.issue_id)
    elif ctx.finish:
        ctx.basket.state_finish(ctx.issue_id)
    elif ctx.accept:
        if ctx.basket.get_state(ctx.issue_id) != 'PROPOSED':
            raise ValueError('Issue already accepted')
        ctx.basket.advance(ctx.issue_id)

    for v in ctx.tag:
        ctx.basket.tag(ctx.issue_id, v)

    for v in ctx.untag:
        ctx.basket.untag(ctx.issue_id, v)

    for v in ctx.unassign:
        ctx.basket.unassign(ctx.issue_id, v)

    for v in ctx.assign:
        ctx.basket.assign(ctx.issue_id, v)
 
    for v in ctx.undep:
        ctx.basket.undep(ctx.issue_id, v)

    for v in ctx.dep:
        ctx.basket.dep(ctx.issue_id, v)

    if ctx.owner:
        ctx.basket.owner(ctx.issue_id, ctx.owner)



# standard imports
import unittest
import logging
import json
from email.message import Message

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.msg import IssueMessage

# test imports
from tests.common import TestStates
from tests.common import TestMsgStore

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


def test_wrapper(p):
    m = Message()
    m.add_header('Foo', 'bar')
    m.set_type('multipart/relative')
    m.set_payload(p)
    return m


def test_unwrapper(msg, message_callback=None, part_callback=None):
    for v in msg.walk():
        if message_callback != None:
            message_callback(v)


class TestMsg(unittest.TestCase):

    def setUp(self):
        self.store = TestStates()
        self.b = Basket(self.store)


    def test_basic(self):
        o = Issue('foo')
        v = IssueMessage(o)


    def test_single_content(self):
        o = Issue('foo')
        v = self.b.add(o)
        m = self.b.msg(v, 's:foo')


    def test_multi_content(self):
        o = Issue('foo')
        v = self.b.add(o)
        m = self.b.msg(v, 's:foo', 's:bar', 's:baz')


    def test_single_file_content(self):
        o = Issue('foo')
        v = self.b.add(o)
        m = self.b.msg(v, 'f:tests/one.png')


    def test_mixed_content(self):
        o = Issue('foo')
        v = self.b.add(o)
        m = self.b.msg(v, 's:bar')
        m = self.b.msg(v, 'f:tests/one.png')
        m = self.b.msg(v, 's:baz')
        m = self.b.msg(v, 'f:tests/two.bin')


    def test_wrapper(self):
        b = Basket(self.store, message_wrapper=test_wrapper)
        o = Issue('bar')
        v = b.add(o)
        m = b.msg(v, 's:foo')
        print(m)


    def test_render(self):
        b = Basket(self.store, message_wrapper=test_wrapper)
        o = Issue('bar')
        v = b.add(o)
        m = b.msg(v, 's:foo')
        m = b.msg(v, 's:bar', 's:baz')
 
        def render_envelope(msg, hdr):
            print('eeeeeenvvv {} {}'.format(hdr, msg))

        def render_message(envelope, msg, mid):
            print('rendeeeer {} {}'.format(mid, msg))

        m = b.get_msg(v, envelope_callback=render_envelope, message_callback=render_message)


    def test_render_alias(self):
        b = Basket(self.store, message_wrapper=test_wrapper)
        o = Issue('bar', alias='xyzzy')
        v = b.add(o)
        m = b.msg(v, 's:foo')
        m = b.msg(v, 's:bar', 's:baz')
 
        def render_envelope(msg, hdr):
            print('eeeeeenvvv {} {}'.format(hdr, msg))

        def render_message(envelope, msg, mid):
            print('rendeeeer {} {}'.format(mid, msg))

        m = b.get_msg('xyzzy', envelope_callback=render_envelope, message_callback=render_message)



if __name__ == '__main__':
    unittest.main()



# standard imports
import unittest
import logging
import json
import shutil
import io
import tempfile
from email.message import Message
from email.utils import localtime as email_localtime

# external imports
from mimeparse import parse_mime_type

# local imports
from piknik import (
        Basket,
        Issue,
        )
from piknik.msg import IssueMessage
from piknik.render.html import Renderer
from piknik.render.html import Accumulator
from piknik.wrap import Wrapper

# test imports
from tests.common import TestStates
from tests.common import TestMsgStore
from tests.common import pgp_setup

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()


def test_wrapper(p):
    m = Message()
    m.add_header('Foo', 'bar')
    m.add_header('X-Piknik-Envelope', 'foo')
    m.set_type('multipart/relative')
    m.set_payload(p)
    return m


class TestMsg(unittest.TestCase):

    def setUp(self):
        (self.crypto, self.gpg, self.gpg_dir) = pgp_setup()
        self.store = TestStates()
        self.b = Basket(self.store, message_wrapper=self.crypto.sign)
        self.render_dir = tempfile.mkdtemp()


    def tearDown(self):
        shutil.rmtree(self.render_dir)


    def test_states_two_issues(self):
        issue_one = Issue('foo')
        self.b.add(issue_one)

        issue_two = Issue('bar')
        v = self.b.add(issue_two)

        m = self.b.msg(v, 's:foo')

        state = self.b.get_state(v)

        msgs = []
        w = io.StringIO()
        renderer = Renderer(self.b, outdir=self.render_dir)
        renderer.apply()


    def test_issue(self):
        issue = Issue('foo')
        issue_id = self.b.add(issue)

        m = self.b.msg(issue_id, 's:foo')

        self.b.tag(issue_id, 'inky')
        self.b.tag(issue_id, 'pinky')

        state = self.b.get_state(issue_id)
        tags = self.b.tags(issue_id)

        w = io.StringIO()
        acc = Accumulator(w=w)
        wrapper = Wrapper()
        renderer = Renderer(self.b, outdir=self.render_dir, wrapper=wrapper, accumulator=acc.add)
        renderer.apply_begin()
        renderer.apply_issue(state, issue, tags)
        renderer.apply_end()

        w.seek(0)
        print(w.read())


    def test_issue_attachment(self):
        b = Basket(self.store, message_wrapper=test_wrapper)
        issue = Issue('foo')
        issue_id = b.add(issue)

        m = b.msg(issue_id, 'f:tests/three.webp')

        state = b.get_state(issue_id)
        tags = []

        w = io.StringIO()
        acc = Accumulator(w=w)
        wrapper = Wrapper()
        renderer = Renderer(b, outdir=self.render_dir, wrapper=wrapper, accumulator=acc.add)
        renderer.apply_begin()
        renderer.apply_issue(state, issue, tags)
        renderer.apply_end()

        w.seek(0)
        print(w.read())



if __name__ == '__main__':
    unittest.main()
